{****************************************************************************}
{*                                                                          *}
{*                              Program/Module                              *}
{*                                   from                                   *}
{*                 NUMERICAL TOOLBOX FOR VERIFIED COMPUTING I               *}
{*                         Basic Numerical Problems                         *}
{*                                                                          *}
{*      Copyright (c) 1993   Rolf Hammer, Matthias Hocks, Dietmar Ratz      *}
{*                                                                          *}
{* This program/module is free software for non-commercial use. For details *}
{* on theory, algorithms, and programs, see the book                        *}
{*                                                                          *}
{*  R. Hammer, M. Hocks, U. Kulisch, D. Ratz:  Numerical Toolbox for        *}
{*  Verified Computing I - Basic Numerical Problems. Springer-Verlag,       *}
{*  Heidelberg (ISBN 3-540-57118-3), New York (ISBN 0-387-57118-3), 1993.   *}
{*                                                                          *}
{* This program/module is distributed WITHOUT ANY WARRANTY. For details,    *}
{* see the "Disclaimer / Legal Matters" of the book (page iv).              *}
{*                                                                          *}
{****************************************************************************}
{----------------------------------------------------------------------------}
{ Purpose: Definition of a multi-dimensional interval differentiation        }
{    arithmetic which allows function evaluation with automatic differen-    }
{    tiation up to second order (i.e. Hessian matrix).                       }
{ Method: Overloading of operators and elementary functions for operations   }
{    of data types 'HessType' and 'HTvector'. Note that all operators and    }
{    functions are implemented using the "modified call by reference" of     }
{    PASCAL-XSC (see Language Reference for details) to avoid the very       }
{    inefficient allocation of local memory for the copies of the actual     }
{    parameters.                                                             }
{ Global types, operators, functions, and procedures:                        }
{    type       HessType, HTvector: data types of differentiation arithmetic }
{    operators  +, -, *, /        : operators of differentiation arithmetic  }
{    operator   :=                : to define 'HessType' constants           }
{    function   HessVar           : to define 'HessType' variables           }
{    functions  fValue, gradValue,                                           }
{               JacValue,                                                    }
{               hessValue         : to get function and derivative values    }
{    functions  sqr, sqrt, power,                                            }
{               exp, sin, cos,... : elementary functions of diff. arithmetic }
{    procedures fEvalH, fEvalJ    : to compute function value only           }
{    procedures fgEvalH, fJEvalJ  : to compute function and first derivative }
{                                   value (gradient or Jacobian)             }
{    procedure  fghEvalH          : to compute function value, gradient, and }
{                                   Hessian matrix value                     }
{----------------------------------------------------------------------------}
module hess_ari;

use i_ari, i_util;         { interval arithmetic, interval utility functions }

{----------------------------------------------------------------------------}
{ Global type definition and local variable                                  }
{----------------------------------------------------------------------------}
global type
  HessType      = dynamic array [*,*] of interval;
                  { The index range must be 0..n,0..n. The component [0,0]   }
                  { contains the function value. The components of the row   }
                  { 0, i.e. [0,1],...,[0,n], contain the value of the        }
                  { gradient. The components [1,1],...,[n,n] contain the     }
                  { value of the Hessian matrix. In fact, only the lower     }
                  { triangle is used (and computed).                         }
  HTvector      = global dynamic array [*] of HessType;
                  { The index range must be 1..n (i.e.: 1..n,0..n,0..n).     }

var                  { The local variable 'HessOrder' is used to select the  }
  HessOrder : 0..2;  { highest order of derivative which is computed. Its    }
                     { default value is 2, and normally the gradient and the }
                     { Hessian matrix are computed.                          }
{----------------------------------------------------------------------------}
{ Transfer operators and functions for constants and variables               }
{----------------------------------------------------------------------------}
global operator := (var u: HessType; x: interval);       { Generate constant }
var i, j: integer;                                       {-------------------}
begin
  u[0,0]:= x;
  for j:=1 to ub(u) do
  begin
    u[0,j]:= 0;
    for i:=1 to ub(u) do u[i,j]:= 0;
  end;
end;

global operator := (var u: HessType; r: real);           { Generate constant }
begin                                                    {-------------------}
  u := intval(r);
end;

global function HessVar (x: ivector)
       : HTvector[1..ub(x)-lb(x)+1, 0..ub(x)-lb(x)+1, 0..ub(x)-lb(x)+1];
var                                                      { Generate variable }
  i, j, k, ubd, d : integer;                             {-------------------}
begin
  ubd := ub(x)-lb(x)+1;   d := 1-lb(x);
  for i:=1 to ubd do
  begin
    HessVar[i,0,0] := x[i-d];  { HessVar[*,0,0]:=x; }
    for k:=1 to ubd do
    begin
      if i=k then HessVar[i,0,k]:= 1 else HessVar[i,0,k]:= 0;
      for j:=1 to ubd do HessVar[i,j,k]:= 0;
    end;
  end;
end;

global function HessVar (v: rvector)
       : HTvector[1..ub(v)-lb(v)+1, 0..ub(v)-lb(v)+1, 0..ub(v)-lb(v)+1];
var                                                      { Generate variable }
  u : ivector[lb(v)..ub(v)];                             {-------------------}
  i : integer;
begin
  for i:=lb(v) to ub(v) do u[i]:= v[i];
  HessVar:= HessVar (u);
end;

{----------------------------------------------------------------------------}
{ Access functions for the function value, the gradient, or the Hessian      }
{----------------------------------------------------------------------------}
global function fValue (var f: HessType) : interval;    { Get function value }
begin                                                   {--------------------}
  fValue:= f[0,0];
end;

global function gradValue (var f: HessType) : ivector[1..ub(f)];
var i: integer;                                         { Get gradient value }
begin                                                   {--------------------}
  for i:=1 to ub(f) do gradValue[i]:= f[0,i];
end;

global function hessValue (var f: HessType) : imatrix[1..ub(f),1..ub(f)];
var i,j: integer;                                       { Get Hessian value  }
begin                                                   {--------------------}
  for i:=1 to ub(f) do
  begin
    hessValue[i,i]:= f[i,i];
    for j:=1 to i-1 do
    begin
      hessValue[i,j]:= f[i,j];  hessValue[j,i]:= f[i,j];
    end;
  end;
end;

{----------------------------------------------------------------------------}
{ Access functions for vector-valued functions (function value or Jacobian)  }
{----------------------------------------------------------------------------}
global function fValue (var f: HTvector) : ivector[1..ub(f,1)];
begin                                                { Get function value of }
  fValue:= f[*,0,0];                                 { n-dimensional vector- }
end;                                                 { valued function       }
                                                     {-----------------------}

global function JacValue (var f: HTvector) : imatrix[1..ub(f,1),1..ub(f,2)];
var i, j: integer;                                   { Get Jacobian value of }
begin                                                { n-dimensional scalar- }
  for i:=1 to ub(f,1) do                             { valued function       }
    for j:=1 to ub(f,2) do                           {-----------------------}
      JacValue[i,j]:= f[i,0,j];
end;

{----------------------------------------------------------------------------}
{ Monadic operators + and - for HessType operands                            }
{----------------------------------------------------------------------------}
global operator + (var u : HessType ) uplus : HessType[0..ub(u),0..ub(u)];
begin
  uplus:= u;
end;

global operator - (var u : HessType ) umin : HessType[0..ub(u),0..ub(u)];
var
  i, j: integer;
begin
  umin[0,0]:= -u[0,0];
  if (HessOrder > 0) then
    for i:=1 to ub(u) do
    begin
      umin[0,i]:= -u[0,i];
      if (HessOrder > 1) then
        for j:=1 to i do umin[i,j]:= -u[i,j];
    end;
end;

{----------------------------------------------------------------------------}
{ Operators +, -, *, and / for two HessType operands                         }
{----------------------------------------------------------------------------}
global operator + (var u,v: HessType ) add : HessType[0..ub(u),0..ub(u)];
var
  i, j: integer;
begin
  add[0,0]:= u[0,0]+v[0,0];
  if (HessOrder > 0) then
    for i:=1 to ub(u) do
    begin
      add[0,i]:= u[0,i]+v[0,i];
      if (HessOrder > 1) then
        for j:=1 to i do add[i,j]:= u[i,j]+v[i,j];
    end;
end;

global operator - (var u,v: HessType ) sub : HessType[0..ub(u),0..ub(u)];
var
  i, j: integer;
begin
  sub[0,0]:= u[0,0]-v[0,0];
  if (HessOrder > 0) then
    for i:=1 to ub(u) do
    begin
      sub[0,i]:= u[0,i]-v[0,i];
      if (HessOrder > 1) then
        for j:=1 to i do sub[i,j]:= u[i,j]-v[i,j];
    end;
end;

global operator * (var u,v: HessType ) mul : HessType[0..ub(u),0..ub(u)];
var
  i, j: integer;
begin
  mul[0,0]:= u[0,0]*v[0,0];
  if (HessOrder > 0) then
    for i:=1 to ub(u) do
    begin
      mul[0,i]:= v[0,0]*u[0,i]+u[0,0]*v[0,i];
      if (HessOrder > 1) then
        for j:=1 to i do
          mul[i,j]:= v[0,0]*u[i,j]+u[0,i]*v[0,j]+v[0,i]*u[0,j]+u[0,0]*v[i,j];
    end;
end;

global operator / (var u,v: HessType ) divis: HessType[0..ub(u),0..ub(u)];
var
  h   : HessType[0..ub(u),0..ub(u)];
  i, j: integer;
begin
  h[0,0]:= u[0,0]/v[0,0];  { Can propagate 'division by zero' error }
  if (HessOrder > 0) then
    for i:=1 to ub(u) do
    begin
      h[0,i]:= (u[0,i]-h[0,0]*v[0,i])/v[0,0];
      if (HessOrder > 1) then
        for j:=1 to i do
          h[i,j]:= (u[i,j]-h[0,i]*v[0,j]-v[0,i]*h[0,j]-h[0,0]*v[i,j])/v[0,0];
    end;
  divis:=h;
end;

{----------------------------------------------------------------------------}
{ Operators +, -, *, and / for one interval and one HessType operand         }
{----------------------------------------------------------------------------}
global operator + (var u: HessType;
                       b: interval) add : HessType[0..ub(u),0..ub(u)];
begin
  add := u;  add[0,0]:= u[0,0]+b;
end;

global operator - (var u: HessType;
                       b: interval) sub : HessType[0..ub(u),0..ub(u)];
begin
  sub := u;  sub[0,0]:= u[0,0]-b;
end;

global operator * (var u: HessType;
                       b: interval) mul : HessType[0..ub(u),0..ub(u)];
var
  i, j: integer;
begin
  mul[0,0]:= u[0,0]*b;
  if (HessOrder > 0) then
    for i:=1 to ub(u) do
    begin
      mul[0,i]:= b*u[0,i];
      if (HessOrder > 1) then
        for j:=1 to i do mul[i,j]:= b*u[i,j];
    end;
end;

global operator / (var u: HessType;
                       b: interval) divis : HessType[0..ub(u),0..ub(u)];
var
  i, j: integer;
begin
  divis[0,0]:= u[0,0]/b;   { Can propagate 'division by zero' error }
  if (HessOrder > 0) then
    for i:=1 to ub(u) do
    begin
      divis[0,i]:= u[0,i]/b;
      if (HessOrder > 1) then
        for j:=1 to i do divis[i,j]:= u[i,j]/b;
    end;
end;

global operator + (    a: interval;
                   var v: HessType) add : HessType[0..ub(v),0..ub(v)];
begin
  add := v;  add[0,0]:= a+v[0,0];
end;

global operator - (    a: interval;
                   var v: HessType) sub : HessType[0..ub(v),0..ub(v)];
var
  i, j: integer;
begin
  sub[0,0]:= a-v[0,0];
  if (HessOrder > 0) then
    for i:=1 to ub(v) do
    begin
      sub[0,i]:= -v[0,i];
      if (HessOrder > 1) then
        for j:=1 to i do sub[i,j]:= -v[i,j];
    end;
end;

global operator * (    a: interval;
                   var v: HessType) mul : HessType[0..ub(v),0..ub(v)];
var
  i, j: integer;
begin
  mul[0,0]:= a*v[0,0];
  if (HessOrder > 0) then
    for i:=1 to ub(v) do
    begin
      mul[0,i]:= a*v[0,i];
      if (HessOrder > 1) then
        for j:=1 to i do mul[i,j]:= a*v[i,j];
    end;
end;

global operator / (    a: interval;
                   var v: HessType) divis : HessType[0..ub(v),0..ub(v)];
var
  h   : HessType[0..ub(v),0..ub(v)];
  i, j: integer;
  p, q: interval;
begin
  h[0,0]:= a/v[0,0];       { Can propagate 'division by zero' error }
  if (HessOrder > 0) then
  begin
    p:= -h[0,0]/v[0,0];  q:= -2*p/v[0,0];
    for i:=1 to ub(v) do
    begin
      h[0,i]:= p*v[0,i];
      if (HessOrder > 1) then
        for j:=1 to i do h[i,j]:= p*v[i,j]+q*v[0,i]*v[0,j];
    end;
  end;
  divis:=h;
end;

{----------------------------------------------------------------------------}
{ Operators +, -, *, and / for one real and one HessType operand             }
{----------------------------------------------------------------------------}
global operator + (var u : HessType;
                       b : real     ) add : HessType[0..ub(u),0..ub(u)];
begin
  add:= u + intval(b);
end;

global operator - (var u: HessType;
                       b: real     ) sub : HessType[0..ub(u),0..ub(u)];
begin
  sub:= u - intval(b);
end;

global operator * (var u: HessType;
                       b: real     ) mul : HessType[0..ub(u),0..ub(u)];
begin
  mul:= u * intval(b);
end;

global operator / (var u: HessType;
                       b: real     ) divs : HessType[0..ub(u),0..ub(u)];
begin
  divs:= u / intval(b);  { Can propagate 'division by zero' error }
end;

global operator + (    a: real;
                   var v: HessType) add : HessType[0..ub(v),0..ub(v)];
begin
  add:= intval(a) + v;
end;

global operator - (    a: real;
                   var v: HessType) sub : HessType[0..ub(v),0..ub(v)];
begin
  sub:= intval(a) - v;
end;

global operator * (    a: real;
                   var v: HessType) mul : HessType[0..ub(v),0..ub(v)];
begin
  mul:= intval(a) * v;
end;

global operator / (    a: real;
                   var v: HessType) divs: HessType[0..ub(v),0..ub(v)];
begin
   divs:= intval(a) / v;  { Can propagate 'division by zero' error }
end;

{----------------------------------------------------------------------------}
{ Elementary functions for HessType arguments                                }
{----------------------------------------------------------------------------}
global function sqr (var u: HessType) : HessType[0..ub(u),0..ub(u)];
var
  i, j: integer;
  h1  : interval;
begin
  sqr[0,0]:= sqr(u[0,0]);
  if (HessOrder > 0) then
  begin
    h1:= 2*u[0,0];
    for i:=1 to ub(u) do
    begin
      sqr[0,i]:= h1*u[0,i];
      if (HessOrder > 1) then
        for j:=1 to i do sqr[i,j]:= h1*u[i,j]+2*u[0,i]*u[0,j];
    end;
  end;
end;

global function power (var u: HessType;
                           k: integer) : HessType[0..ub(u),0..ub(u)];
var
  h1   : interval;
  i, j : integer;
begin
  if (k = 0) then
    power:= 1
  else if (k = 1) then
    power:= u
  else if (k = 2) then
    power:= sqr(u)
  else
    begin
      power[0,0]:= power(u[0,0], k);
      if (HessOrder > 0) then
      begin
        h1:= k * power(u[0,0], k-1);
        for i:=1 to ub(u) do
        begin
          power[0,i]:= h1 * u[0,i];
          if (HessOrder > 1) then
            for j:=1 to i do
              power[i,j]:= h1*u[i,j] + k*(k-1)*power(u[0,0],k-2)*u[0,i]*u[0,j]
        end;
      end;
    end;
end;

global function sqrt (var u: HessType) : HessType[0..ub(u),0..ub(u)];
var
  i, j       : integer;
  h0, h1, h2 : interval;
begin
  h0:= sqrt(u[0,0]);  { Can propagate domain error }
  sqrt[0,0]:= h0;
  if (HessOrder > 0) then
  begin
    h1:= 0.5/h0;  h2:= -0.5*h1/u[0,0];
    for i:=1 to ub(u) do
    begin
      sqrt[0,i]:= h1*u[0,i];
      if (HessOrder > 1) then
        for j:=1 to i do sqrt[i,j]:= h1*u[i,j]+h2*u[0,i]*u[0,j];
    end;
  end;
end;

global function exp (var u: HessType) : HessType[0..ub(u),0..ub(u)];
var
  h0  : interval;
  i, j: integer;
begin
  h0:= exp(u[0,0]);  exp[0,0]:= h0;
  if (HessOrder > 0) then
    for i:=1 to ub(u) do
    begin
      exp[0,i]:= h0*u[0,i];
      if (HessOrder > 1) then
        for j:=1 to i do exp[i,j]:= h0*(u[i,j]+u[0,i]*u[0,j]);
    end;
end;

global function ln (var u: HessType) : HessType[0..ub(u),0..ub(u)];
var
  i, j   : integer;
  h1, h2 : interval;
begin
  ln[0,0]:= ln(u[0,0]);    { Can propagate domain error }
  if (HessOrder > 0) then
  begin
    h1:= 1/u[0,0];  h2:= -sqr(h1);
    for i:=1 to ub(u) do
    begin
      ln[0,i]:= h1*u[0,i];
      if (HessOrder > 1) then
        for j:=1 to i do ln[i,j]:= h1*u[i,j]+h2*u[0,i]*u[0,j];
    end;
  end;
end;

global function sin (var u: HessType) : HessType[0..ub(u),0..ub(u)];
var
  i, j       : integer;
  h0, h1, h2 : interval;
begin
  h0:= sin(u[0,0]);  sin[0,0]:= h0;
  if (HessOrder > 0) then
  begin
    h1:= cos(u[0,0]);  h2:= -h0;
    for i:=1 to ub(u) do
    begin
      sin[0,i]:= h1*u[0,i];
      if (HessOrder > 1) then
        for j:=1 to i do sin[i,j]:= h1*u[i,j]+h2*u[0,i]*u[0,j];
    end;
  end;
end;

global function cos (var u: HessType) : HessType[0..ub(u),0..ub(u)];
var
 i, j       : integer;
 h0, h1, h2 : interval;
begin
  h0:= cos(u[0,0]);  cos[0,0]:= h0;
  if (HessOrder > 0) then
  begin
    h1:= -sin(u[0,0]);  h2:= -h0;
    for i:=1 to ub(u) do
    begin
      cos[0,i]:= h1*u[0,i];
      if (HessOrder > 1) then
        for j:=1 to i do cos[i,j]:= h1*u[i,j]+h2*u[0,i]*u[0,j];
    end;
  end;
end;

global function tan (var u: HessType) : HessType[0..ub(u),0..ub(u)];
var
 i, j       : integer;
 h0, h1, h2 : interval;
begin
  h0:= tan(u[0,0]);  { Can propagate domain error }
  tan[0,0]:= h0;
  if (HessOrder > 0) then
  begin                                   { The subdistributive law implies  }
    h1:= sqr(h0)+1;  h2:= 2*h0*h1;        {   h0 * (h0^2 + 1) <= h0^3 + h0   }
    for i:=1 to ub(u) do                  { So, we use the first form.       }
    begin
      tan[0,i] := h1*u[0,i];
      if (HessOrder > 1) then
        for j:=1 to i do tan[i,j]:= h1*u[i,j]+h2*u[0,i]*u[0,j];
    end;
  end;
end;

global function cot (var u: HessType) : HessType[0..ub(u),0..ub(u)];
var
 i, j       : integer;
 h0, h1, h2 : interval;
begin
  h0:= cot(u[0,0]);  { Can propagate domain error }
  cot[0,0]:= h0;
  if (HessOrder > 0) then
  begin                                   { The subdistributive law implies  }
    h1:= -(sqr(h0)+1);  h2:= -2*h0*h1;    {   h0 * (h0^2 + 1) <= h0^3 + h0   }
    for i:=1 to ub(u) do                  { So, we use the first form.       }
    begin
      cot[0,i] := h1*u[0,i];
      if (HessOrder > 1) then
        for j:=1 to i do cot[i,j]:= h1*u[i,j]+h2*u[0,i]*u[0,j];
    end;
  end;
end;

global function arcsin (var u: HessType) : HessType[0..ub(u),0..ub(u)];
var
 i, j      : integer;
 h, h1, h2 : interval;
begin
  arcsin[0,0]:= arcsin(u[0,0]);  { Can propagate domain error }
  if (HessOrder > 0) then
  begin
    h := 1-sqr(u[0,0]);  h1:= 1/sqrt(h);  h2:= u[0,0]*h1/h;
    for i:=1 to ub(u) do
    begin
      arcsin[0,i]:= h1*u[0,i];
      if (HessOrder > 1) then
      for j:=1 to i do arcsin[i,j]:= h1*u[i,j]+h2*u[0,i]*u[0,j];
    end;
  end;
end;

global function arccos (var u: HessType) : HessType[0..ub(u),0..ub(u)];
var
 i, j      : integer;
 h, h1, h2 : interval;
begin
  arccos[0,0]:= arccos(u[0,0]);  { Can propagate domain error }
  if (HessOrder > 0) then
  begin
    h := 1-sqr(u[0,0]);  h1:= -1/sqrt(h);  h2:= u[0,0]*h1/h;
    for i:=1 to ub(u) do
    begin
      arccos[0,i]:= h1*u[0,i];
      if (HessOrder > 1) then
        for j:=1 to i do arccos[i,j]:= h1*u[i,j]+h2*u[0,i]*u[0,j];
    end;
  end;
end;

global function arctan (var u: HessType) : HessType[0..ub(u),0..ub(u)];
var
 i, j   : integer;
 h1, h2 : interval;
begin
  arctan[0,0]:= arctan(u[0,0]);  { Can propagate domain error }
  if (HessOrder > 0) then
  begin
    h1:= 1/(1+sqr(u[0,0]));  h2:= -2*u[0,0]*sqr(h1);
    for i:=1 to ub(u) do
    begin
      arctan[0,i]:= h1*u[0,i];
      if (HessOrder > 1) then
        for j:=1 to i do arctan[i,j]:= h1*u[i,j]+h2*u[0,i]*u[0,j];
    end;
  end;
end;

global function arccot (var u: HessType) : HessType[0..ub(u),0..ub(u)];
var
 i, j   : integer;
 h1, h2 : interval;
begin
  arccot[0,0]:= arccot(u[0,0]);  { Can propagate domain error }
  if (HessOrder > 0) then
  begin
    h1:= -1/(1+sqr(u[0,0]));  h2:= 2*u[0,0]*sqr(h1);
    for i:=1 to ub(u) do
    begin
      arccot[0,i]:= h1*u[0,i];
      if (HessOrder > 1) then
        for j:=1 to i do arccot[i,j]:= h1*u[i,j]+h2*u[0,i]*u[0,j];
    end;
  end;
end;

global function sinh (var u: HessType) : HessType[0..ub(u),0..ub(u)];
var
  i, j       : integer;
  h0, h1, h2 : interval;
begin
  h0:= sinh(u[0,0]);  sinh[0,0]:= h0;
  if (HessOrder > 0) then
  begin
    h1:= cosh(u[0,0]);  h2:= h0;
    for i:=1 to ub(u) do
    begin
      sinh[0,i]:= h1*u[0,i];
      if (HessOrder > 1) then
        for j:=1 to i do sinh[i,j]:= h1*u[i,j]+h2*u[0,i]*u[0,j];
    end;
  end;
end;

global function cosh (var u: HessType) : HessType[0..ub(u),0..ub(u)];
var
 i, j       : integer;
 h0, h1, h2 : interval;
begin
  h0:= cosh(u[0,0]);  cosh[0,0]:= h0;
  if (HessOrder > 0) then
  begin
    h1:= sinh(u[0,0]);  h2:= h0;
    for i:=1 to ub(u) do
    begin
      cosh[0,i] := h1*u[0,i];
      if (HessOrder > 1) then
        for j:=1 to i do cosh[i,j]:= h1*u[i,j]+h2*u[0,i]*u[0,j];
    end;
  end;
end;

global function tanh (var u: HessType) : HessType[0..ub(u),0..ub(u)];
var
 i, j       : integer;
 h0, h1, h2 : interval;
begin
  h0:= tanh(u[0,0]);  tanh[0,0]:= h0;
  if (HessOrder > 0) then
  begin
    h1:= 1-sqr(h0);  h2:= -2*h0*h1;       { The subdistributive law implies  }
    for i:=1 to ub(u) do                  {   h0 * (h0^2 - 1) <= h0^3 - h0   }
    begin                                 { So, we use the first form.       }
      tanh[0,i] := h1*u[0,i];
      if (HessOrder > 1) then
        for j:=1 to i do tanh[i,j]:= h1*u[i,j]+h2*u[0,i]*u[0,j];
    end;
  end;
end;

global function coth (var u: HessType) : HessType[0..ub(u),0..ub(u)];
var
 i, j       : integer;
 h0, h1, h2 : interval;
begin
  h0:= coth(u[0,0]);  { Can propagate domain error }
  coth[0,0]:= h0;
  if (HessOrder > 0) then
  begin                                   { The subdistributive law implies  }
    h1:= 1-sqr(h0);  h2:= -2*h0*h1;       {   h0 * (h0^2 - 1) <= h0^3 - h0   }
    for i:=1 to ub(u) do                  { So, we use the first form.       }
    begin
      coth[0,i] := h1*u[0,i];
      if (HessOrder > 1) then
        for j:=1 to i do coth[i,j]:= h1*u[i,j]+h2*u[0,i]*u[0,j];
    end;
  end;
end;

global function arsinh (var u: HessType) : HessType[0..ub(u),0..ub(u)];
var
 i, j      : integer;
 h, h1, h2 : interval;
begin
  arsinh[0,0]:= arsinh(u[0,0]);  { Can propagate domain error }
  if (HessOrder > 0) then
  begin
    h:= 1+sqr(u[0,0]);  h1:= 1/sqrt(h);  h2:= -u[0,0]*h1/h;
    for i:=1 to ub(u) do
    begin
      arsinh[0,i]:= h1*u[0,i];
      if (HessOrder > 1) then
        for j:=1 to i do arsinh[i,j]:= h1*u[i,j]+h2*u[0,i]*u[0,j];
    end;
  end;
end;

global function arcosh (var u: HessType) : HessType[0..ub(u),0..ub(u)];
var
  i, j      : integer;
  h, h1, h2 : interval;
begin
  arcosh[0,0]:= arcosh(u[0,0]);  { Can propagate domain error }
  if (HessOrder > 0) then
  begin
    h:= sqr(u[0,0])-1;  h1:= 1/sqrt(h);  h2:= -u[0,0]*h1/h;
    for i:=1 to ub(u) do
    begin
      arcosh[0,i]:= h1*u[0,i];
      if (HessOrder > 1) then
        for j:=1 to i do arcosh[i,j]:= h1*u[i,j]+h2*u[0,i]*u[0,j];
    end;
  end;
end;

global function artanh (var u: HessType) : HessType[0..ub(u),0..ub(u)];
var
 i, j   : integer;
 h1, h2 : interval;
begin
  artanh[0,0]:= artanh(u[0,0]);  { Can propagate domain error }
  if (HessOrder > 0) then
  begin
    h1:= 1/(1-sqr(u[0,0]));  h2:= 2*u[0,0]*sqr(h1);
    for i:=1 to ub(u) do
    begin
      artanh[0,i]:= h1*u[0,i];
      if (HessOrder > 1) then
        for j:=1 to i do artanh[i,j]:= h1*u[i,j]+h2*u[0,i]*u[0,j];
    end;
  end;
end;

global function arcoth (var u: HessType) : HessType[0..ub(u),0..ub(u)];
var
 i, j   : integer;
 h1, h2 : interval;
begin
  arcoth[0,0]:= arcoth(u[0,0]);  { Can propagate domain error }
  if (HessOrder > 0) then
  begin
    h1:= 1/(1-sqr(u[0,0]));  h2:= 2*u[0,0]*sqr(h1);
    for i:=1 to ub(u) do
    begin
      arcoth[0,i]:= h1*u[0,i];
      if (HessOrder > 1) then
        for j:=1 to i do arcoth[i,j]:= h1*u[i,j]+h2*u[0,i]*u[0,j];
    end;
  end;
end;


{----------------------------------------------------------------------------}
{ Predefined routines for evaluation of HessType-functions                   }
{----------------------------------------------------------------------------}
{ Purpose: Evaluation of function 'f' for argument 'x' in differentiation    }
{    arithmetic computing only the function value.                           }
{ Parameters:                                                                }
{    In     : 'f'       : function of 'HessType'.                            }
{             'x'       : argument for evaluation of 'f'.                    }
{    Out    : 'fx'      : returns the function value 'f(x)'.                 }
{ Description: This procedure sets 'HessOrder' to 0, evaluates 'f(x)' in     }
{    differentiation arithmetic, and returns the function value only.        }
{----------------------------------------------------------------------------}
global procedure fEvalH (function f(y:HTvector) : HessType[lb(y,2)..ub(y,2),
                                                           lb(y,3)..ub(y,3)];
                                              x : ivector;
                              var            fx : interval);
begin
  HessOrder:= 0;  fx:= fValue(f(HessVar(x)));  HessOrder:= 2;
end;

{----------------------------------------------------------------------------}
{ Purpose: Evaluation of function 'f' for argument 'x' in differentiation    }
{    arithmetic computing the function value and the gradient value.         }
{ Parameters:                                                                }
{    In     : 'f'       : function of 'HessType'.                            }
{             'x'       : argument for evaluation of 'f'.                    }
{    Out    : 'fx'      : returns the function value 'f(x)'.                 }
{             'gx'      : returns the gradient value 'grad f(x)'.            }
{ Description: This procedure sets 'HessOrder' to 1, evaluates 'f(x)' in     }
{    differentiation arithmetic, and returns the function value and the      }
{    value of the gradient.                                                  }
{----------------------------------------------------------------------------}
global procedure fgEvalH (function f(y:HTvector) : HessType[lb(y,2)..ub(y,2),
                                                            lb(y,3)..ub(y,3)];
                                               x : ivector;
                               var            fx : interval;
                               var            gx : ivector);
var
  fxH : HessType[0..ub(x),0..ub(x)];
begin
  HessOrder:= 1;
  fxH:= f(HessVar(x));  fx:= fValue(fxH);  gx:= gradValue(fxH);
  HessOrder := 2;
end;

{----------------------------------------------------------------------------}
{ Purpose: Evaluation of function 'f' for argument 'x' in differentiation    }
{    arithmetic computing the function value, the gradient value, and the    }
{    Hessian matrix value.                                                   }
{ Parameters:                                                                }
{    In     : 'f'       : function of 'HessType'.                            }
{             'x'       : argument for evaluation of 'f'.                    }
{    Out    : 'fx'      : returns the function value 'f(x)'.                 }
{             'gx'      : returns the gradient value 'grad f(x)'.            }
{             'hx'      : returns the Hessian matrix value 'hess f(x)'.      }
{ Description: This procedure keeps 'HessOrder' = 2, evaluates 'f(x)' in     }
{    differentiation arithmetic, and returns the function value, the value   }
{    of the gradient, and the value of the Hessian matrix.                   }
{----------------------------------------------------------------------------}
global procedure fghEvalH (function f(y:HTvector): HessType[lb(y,2)..ub(y,2),
                                                            lb(y,3)..ub(y,3)];
                                               x : ivector;
                                var           fx : interval;
                                var           gx : ivector;
                                var           hx : imatrix);
var
  fxH : HessType[0..ub(x),0..ub(x)];
begin
  fxH:= f(HessVar(x));
  fx:= fValue(fxH);  gx:= gradValue(fxH);  hx:= hessValue(fxH);
end;

{----------------------------------------------------------------------------}
{ Predefined routines for evaluation of HTvector-functions (with Jacobians)  }
{----------------------------------------------------------------------------}
{ Purpose: Evaluation of function 'f' for argument 'x' in differentiation    }
{    arithmetic computing only the function value.                           }
{ Parameters:                                                                }
{    In     : 'f'       : function of type 'HTvector'.                       }
{             'x'       : argument for evaluation of 'f'.                    }
{    Out    : 'fx'      : returns the vector function value 'f(x)'.          }
{ Description: This procedure sets 'HessOrder' to 0, evaluates 'f(x)' in     }
{    differentiation arithmetic, and returns the function value only.        }
{----------------------------------------------------------------------------}
global procedure fEvalJ (function f(y:HTvector) : HTvector[lb(y,1)..ub(y,1),
                                                           lb(y,2)..ub(y,2),
                                                           lb(y,3)..ub(y,3)];
                                              x : ivector;
                              var            fx : ivector);
begin
  HessOrder:= 0;  fx:= fValue(f(HessVar(x)));  HessOrder:= 2;
end;

{----------------------------------------------------------------------------}
{ Purpose: Evaluation of function 'f' for argument 'x' in differentiation    }
{    arithmetic computing the function value and the Jacobian matrix value.  }
{ Parameters:                                                                }
{    In     : 'f'       : function of type 'HTvector'.                       }
{             'x'       : argument for evaluation of 'f'.                    }
{    Out    : 'fx'      : returns the function value 'f(x)'.                 }
{             'Jx'      : returns the Jacobian value 'Jf(x)'.                }
{ Description: This procedure sets 'HessOrder' to 1, evaluates 'f(x)' in     }
{    differentiation arithmetic, and returns the function value and the      }
{    value of the Jacobian matrix.                                           }
{----------------------------------------------------------------------------}
global procedure fJEvalJ (function f(y:HTvector) : HTvector[lb(y,1)..ub(y,1),
                                                            lb(y,2)..ub(y,2),
                                                            lb(y,3)..ub(y,3)];
                                               x : ivector;
                               var            fx : ivector;
                               var            Jx : imatrix);
var
  fxGTv : HTvector[1..ub(x),0..ub(x),0..ub(x)];
begin
  HessOrder:= 1;
  fxGTv:= f(HessVar(x));  fx:= fValue(fxGTv);  Jx:= JacValue(fxGTv);
  HessOrder:= 2;
end;

{----------------------------------------------------------------------------}
{ Module initialization part                                                 }
{----------------------------------------------------------------------------}
begin
  HessOrder := 2;
end.
