{****************************************************************************}
{*                                                                          *}
{*                              Program/Module                              *}
{*                                   from                                   *}
{*                 NUMERICAL TOOLBOX FOR VERIFIED COMPUTING I               *}
{*                         Basic Numerical Problems                         *}
{*                                                                          *}
{*      Copyright (c) 1993   Rolf Hammer, Matthias Hocks, Dietmar Ratz      *}
{*                                                                          *}
{* This program/module is free software for non-commercial use. For details *}
{* on theory, algorithms, and programs, see the book                        *}
{*                                                                          *}
{*  R. Hammer, M. Hocks, U. Kulisch, D. Ratz:  Numerical Toolbox for        *}
{*  Verified Computing I - Basic Numerical Problems. Springer-Verlag,       *}
{*  Heidelberg (ISBN 3-540-57118-3), New York (ISBN 0-387-57118-3), 1993.   *}
{*                                                                          *}
{* This program/module is distributed WITHOUT ANY WARRANTY. For details,    *}
{* see the "Disclaimer / Legal Matters" of the book (page iv).              *}
{*                                                                          *}
{****************************************************************************}
{----------------------------------------------------------------------------}
{ Purpose: Definition of a linearly linked list as an abstract data type for }
{    the representation of a list of integer sets.                           }
{ Global constants, types, operators, functions, and procedures:             }
{    constant maxDim      : maximum number of indices in a set (restricted   }
{                           by the PASCAL-XSC compiler)                      }
{    type IndexSet        : representation of a set of indices               }
{    type BaseList        : representation of a linearly linked list of      }
{                           index sets                                       }
{    operator in          : returns TRUE, if index set is in list            }
{    operator :=          : assigns index set to a vector                    }
{    function NextIndex   : returns value of next index in an index set      }
{    function GetIndex    : returns value of i-th index in an index set      }
{    function size        : returns number of indices in an index set        }
{    function length      : returns number of elements in a list             }
{    function empty       : returns TRUE if list is empty                    }
{    function select      : selects first element of a list                  }
{    function extract     : extracts submatrix/subvector depending on the    }
{                           actual index set                                 }
{    procedure insert     : inserts index set at the head of a list          }
{    procedure delete     : deletes index set from list                      }
{    procedure append     : appends 2nd list to end of 1st list              }
{    procedure remove     : removes elements of 2nd list from 1st list       }
{    procedure FreeAll    : frees complete list                              }
{    procedure write      : prints data types IndexSet and BaseList          }
{----------------------------------------------------------------------------}
module lop_ari;

global const    maxDim = 255; { Maximum number of indices }

{----------------------------------------------------------------------------}
{ Global type definitions                                                    }
{----------------------------------------------------------------------------}

global type     IndexSet = set of 1..maxDim;
                BaseList = ^BaseListElement;
                BaseListElement = 
                          record
                            Member : IndexSet;
                            next   : ^BaseListElement;
                          end;

{----------------------------------------------------------------------------}
{ Local variable storing list of free elements (automatic garbage recycling) }
{----------------------------------------------------------------------------}
var
  FreeList : BaseList;

{----------------------------------------------------------------------------}
{ Procedures for generating and freeing of list elements                     }
{----------------------------------------------------------------------------}
procedure NewPP (var P: BaseList);    { 'NewPP' generates a new list element }
begin                                 { or gets one from 'FreeList'.         }
  if FreeList = nil then              {--------------------------------------}
    begin
      new(P);  P^.next:= nil;
    end
  else
    begin
      P:= FreeList;  FreeList:= FreeList^.next;  P^.next:= nil;
    end;
end;

procedure Free (var P: BaseList);           { 'Free' enters one element of a }
begin                                       { list in the 'FreeList'.        }
  if P <> nil then                          {--------------------------------}
  begin
      P^.next:= FreeList;  FreeList := P;  P:= nil;
  end;
end;

global procedure FreeAll (var List: BaseList);{ 'FreeAll' enters all elements}
var  H : BaseList;                            { of 'List' in the 'FreeList'. }
begin                                         {------------------------------}
  if List <> nil then
  begin
    H:= List;
    while H^.next <> nil do  H:= H^.next;
    H^.next:= FreeList;  FreeList:= List;  List:= nil;
  end;
end;

{----------------------------------------------------------------------------}
{ Global functions and procedures for index sets                             }
{----------------------------------------------------------------------------}
global function NextIndex(Index : integer;            { Determine next index }
                              B : IndexSet) : integer;{ in B following Index }
begin                                                 {----------------------}
  Index := Index + 1;
  while not (Index in B) and (Index < maxDim) do Index := Index + 1;
  NextIndex := Index;
end;

global function GetIndex(i : integer;            { Determine i-th index in B }
                         B : IndexSet) : integer;{---------------------------}
var Index, Counter : integer;

begin
  Index := 0; Counter := 0;
  repeat
    Index := NextIndex(Index, B);
    Counter := Counter+1;
  until (Counter=i);
  GetIndex := Index;
end;

global function size(B : IndexSet) : integer; { Determine number of indices  }
var Index, k : integer;                       { in B                         }
begin                                         {------------------------------}
  k := 0;
  for Index:=1 to maxDim do
    if (Index in B) then
      k := k+1;
  size := k;
end;
 
global function extract(A : rmatrix; B : IndexSet) : { Extract submatrix of  }
                     rmatrix[1..ub(A,1),1..size(B)]; { columns A[*,Ind] with }
var     Index, k : integer;                          { (Index in B)          }
begin                                                {-----------------------}
  k := 0;
  for Index:=1 to ub(A,2) do
    if (Index in B) then
      begin
        k := k+1;
        extract[*,k] := A[*,Index];
      end;
end;

global function extract(x : rvector; B : IndexSet) : { Extract subvector     }
                                rvector[1..size(B)]; { x[Index] with (Index  }
var     Index, k : integer;                          { in B)                 }
begin                                                {-----------------------}
  k := 0;
  for Index:=1 to ub(x) do
    if (Index in B) then
      begin
        k := k+1;
        extract[k] := x[Index];
      end;
end;

{----------------------------------------------------------------------------}
{ Global functions and procedures for lists of index sets                    }
{----------------------------------------------------------------------------}
global function empty(L : BaseList) : boolean;         { Check if L is empty }
begin                                                  {---------------------}
  empty := (L = nil);
end;

global function length(L : BaseList) : integer;       { Return length of L   }
var counter : integer;                                {----------------------}
begin
  counter := 0;
  while (L <> nil) do
    begin
      counter := counter + 1;
      L := L^.next;
    end;
 length := counter;
end;

global operator in (B : IndexSet; L : BaseList)            { Check if B in L }
                              inlist : boolean;            {-----------------}
begin
  inlist := false;
  while (L <> nil) and (not inlist) do
    if (L^.Member = B) then inlist := true
                             else L := L^.next;
end;

global function select(L : BaseList) : IndexSet;    { Select first index set } 
begin                                               { from L                 }
  select := L^.Member;                              {------------------------}
end;

global procedure insert (var L : BaseList;  { Insert index set B to the head }
                             B : IndexSet); { of the list of index sets L    }
var P : BaseList;                           {--------------------------------}
begin
  if not (B in L) then
    begin
      P := L;
      NewPP(L);
      L^.Member := B;
      L^.next := P;
    end;
end;

global procedure delete (var L : BaseList;  { Delete index set B from list   }
                             B : IndexSet); { of index sets L                }
var P, Del : BaseList;                      {--------------------------------}
begin
  if not empty(L) then 
    if (L^.Member = B) then { B is 1st element }
      begin
        Del := L; L := L^.next; Free(Del);
      end
    else { B is not 1st element }
      begin
        P := L;
        while (P^.next <> nil) do
          if (P^.next^.Member = B) then 
            begin
              Del := P^.next; P^.next := P^.next^.next; Free(Del);
            end
          else P := P^.next;
      end;
end;

global procedure append (var Res : BaseList;  { Append list of index sets Add}
                         var Add : BaseList); { to list Res                  }
var P : BaseList;                             {------------------------------}
begin
  if empty(Res) then 
    Res := Add
  else
    begin
      P := Res;
      while (P^.next <> nil) do P := P^.next;
      P^.next := Add;
    end;
  Add := nil;
end;

global procedure remove (var Res : BaseList;  { Remove list of index sets Sub}
                             Sub : BaseList); { from Res                     }
var B : IndexSet;                             {------------------------------}
begin
  while (Sub <> nil) do
    begin
      B := Sub^.Member;
      delete(Res,B);
      Sub := Sub^.next;
    end;
end;

global procedure write(var out : text; B : IndexSet);      { Write index set }
var Index : integer;                                       {-----------------}
begin
  for Index:=1 to maxDim do
    if (Index in B) then write(out,Index,' ');
end;

global procedure write(var out : text;            { Write list of index sets }
                       BaseSet : BaseList);       {--------------------------}
begin
  while (BaseSet <> nil) do
    begin
      writeln(out,BaseSet^.Member);
      BaseSet := BaseSet^.next;
    end;
end;

global operator := (var x : rvector; B : IndexSet);    { Assign index set to }
var i, Index : integer;                                { real vector         }
begin                                                  {---------------------}
  Index := 0;
  for i:=1 to ub(x) do
    begin
      Index := NextIndex(Index,B);
      x[i] := Index;
    end;
end;

{----------------------------------------------------------------------------}
{ Module initialization                                                      }
{----------------------------------------------------------------------------}
begin
  FreeList := nil;         { List of freed elements which can be used again  }
end.                       {-------------------------------------------------}
