{****************************************************************************}
{*                                                                          *}
{*                              Program/Module                              *}
{*                                   from                                   *}
{*                 NUMERICAL TOOLBOX FOR VERIFIED COMPUTING I               *}
{*                         Basic Numerical Problems                         *}
{*                                                                          *}
{*      Copyright (c) 1993   Rolf Hammer, Matthias Hocks, Dietmar Ratz      *}
{*                                                                          *}
{* This program/module is free software for non-commercial use. For details *}
{* on theory, algorithms, and programs, see the book                        *}
{*                                                                          *}
{*  R. Hammer, M. Hocks, U. Kulisch, D. Ratz:  Numerical Toolbox for        *}
{*  Verified Computing I - Basic Numerical Problems. Springer-Verlag,       *}
{*  Heidelberg (ISBN 3-540-57118-3), New York (ISBN 0-387-57118-3), 1993.   *}
{*                                                                          *}
{* This program/module is distributed WITHOUT ANY WARRANTY. For details,    *}
{* see the "Disclaimer / Legal Matters" of the book (page iv).              *}
{*                                                                          *}
{****************************************************************************}
{----------------------------------------------------------------------------}
{ Purpose: Definition of a list arithmetic used in connection with an        }
{    interval bisection method in global optimization for storing pairs of   }
{    an interval and a real value.                                           }
{ Method: Overloading of functions and operators of data types 'Pair' (list  }
{    element) and 'PairPtr' (list).                                          }
{ Global types, operators, functions, and procedures:                        }
{    types      Pair             : list elements (pair of interval and real) }
{               PairPtr, PairElmt: list of pairs                             }
{    operators  +                : adding a new element to a list            }
{    functions  MakePair         : transfer function for pairs               }
{               GetInt, GetFyi   : access functions for pairs                }
{               Next, Head       : access functions for lists                }
{               Length           : access function to length of list         }
{               EmptyList        : delivers an empty list                    }
{    procedures FreeAll          : free complete list                        }
{               MultiDelete      : deletes several elements in a list        }
{               DelHead          : deletes the first element of a list       }
{----------------------------------------------------------------------------}
module lst1_ari;

use
  i_ari;   { Interval arithmetic }

{----------------------------------------------------------------------------}
{ Global type definitions                                                    }
{----------------------------------------------------------------------------}
global type
  Pair     = record
               int : interval;               { Pair of an interval 'y' and   }
               fyi : real;                   { the real value 'inf(f(y))'.   }
             end;                            {-------------------------------}
  PairPtr  = ^PairElmt;
  PairElmt = record                          { List of pairs }
               P : Pair;                     {---------------}
               N : PairPtr;
             end;

{----------------------------------------------------------------------------}
{ Local variable storing list of free elements (automatic garbage recycling) }
{----------------------------------------------------------------------------}
var
  FreeList : PairPtr;

{----------------------------------------------------------------------------}
{ Procedures for generating and freeing of list elements (pairs)             }
{----------------------------------------------------------------------------}
procedure NewPP (var pp: PairPtr);    { 'NewPP' generates a new list element }
begin                                 { or gets one from 'FreeList'.         }
  if FreeList = nil then              {--------------------------------------}
    begin
      new(pp);  pp^.N:= nil;
    end
  else
    begin
      pp:= FreeList;  FreeList:= FreeList^.N;  pp^.N:= nil;
    end;
end;

procedure Free (var pp: PairPtr);           { 'Free' enters one element of a }
begin                                       { list in the 'FreeList'.        }
  if pp <> nil then                         {--------------------------------}
  begin
      pp^.N:= FreeList;  FreeList := pp;  pp:= nil;
  end;
end;

global procedure FreeAll (var List: PairPtr);{ 'FreeAll' enters all elements }
var  H : PairPtr;                            { of 'List' in the 'FreeList'.  }
begin                                        {-------------------------------}
  if List <> nil then
  begin
    H:= List;
    while H^.N <> nil do  H:= H^.N;
    H^.N:= FreeList;  FreeList:= List;  List:= nil;
  end;
end;

{----------------------------------------------------------------------------}
{ Transfer and access functions for pairs                                    }
{----------------------------------------------------------------------------}
global function MakePair (int: interval; fyi: real) : Pair;  { Generate pair }
begin                                                        {---------------}
  MakePair.int:= int;  MakePair.fyi:= fyi;
end;

global function GetInt (P: Pair) : interval;             { Get int-component }
begin                                                    {-------------------}
  GetInt:= P.int;
end;

global function GetFyi (P: Pair) : real;                 { Get fyi-component }
begin                                                    {-------------------}
  GetFyi:= P.fyi;
end;

{----------------------------------------------------------------------------}
{ Operators, functions, and procedures for lists of pairs                    }
{----------------------------------------------------------------------------}
{ Global function 'EmptyList' representing an empty list of pairs.           }
{----------------------------------------------------------------------------}
global function EmptyList : PairPtr;
begin
  EmptyList := nil;
end;

{----------------------------------------------------------------------------}
{ Operator + enters the pair 'P' the list 'List' in such a way that after    }
{ entering, one of the four following condition holds:                       }
{   1) O.fyi <= P.fyi < Q.fyi,                                               }
{   2)          P.fyi < Q.fyi  and  'P' is the first element of 'List',      }
{   3) O.fyi <= P.fyi          and  'P' is the last  element of 'List',      }
{   4)                              'P' is the only  element of 'List',      }
{ where 'O' is the preceding and 'Q' is the succeeding element of 'P' in     }
{ the resulting list.                                                        }
{----------------------------------------------------------------------------}
global operator + (List: PairPtr; P: Pair) Enter: PairPtr;
var
  H, HN            : PairPtr;
  ready, alreadyIn : boolean;
begin
  if (List = nil) then                               { List is empty, so new }
    begin                                            { list contains only P. }
      NewPP(H);  H^.P:= P;  H^.N:= nil;  Enter:= H;  {-----------------------}
    end
  else if (List^.P.fyi > P.fyi) then                 { P becomes new first   }
    begin                                            { element of the list.  }
      NewPP(H);  H^.P:= P;  H^.N:= List;  Enter:= H; {-----------------------}
    end
  else
    begin
      H:= List;  HN:= H^.N;  ready:= false;  alreadyIn:= H^.P.int = P.int;

      while not (ready or alreadyIn) do              { Search for the right  }
      begin                                          { position to enter P.  }
        if (HN = nil) then                           {-----------------------}
          ready:= true
        else if (HN^.P.fyi > P.fyi) then
          ready:= true
        else
          begin
            H:= HN;  HN:= H^.N;  alreadyIn:= H^.P.int = P.int;
          end;
      end;

      if not alreadyIn then
      begin
        NewPP(H^.N);  H:= H^.N;                      { Enter P between H and }
        H^.P:= P;  H^.N:= HN;                        { HN. Return List.      }
      end;                                           {-----------------------}

      Enter:= List;
    end;
end;

{----------------------------------------------------------------------------}
{ 'MultiDelete' deletes all elements 'P' in 'List' for which the condition   }
{ 'P.fyi > fmax' holds.  This procedure assumes that the 'fyi' components of }
{ the list elements are sorted in increasing order (see operator +).         }
{----------------------------------------------------------------------------}
global procedure MultiDelete (var List: PairPtr; fmax: real);
var
  DelPrev, Del : PairPtr;
  ready        : boolean;
begin
  if (List <> nil) then
  begin
    if (List^.P.fyi > fmax) then                 { All list elements satisfy }
      begin                                      { 'P.fyi > fmax'.           }
        Del:= List;  List:= nil;                 {---------------------------}
      end
    else
      begin
        DelPrev:= List;  Del:= DelPrev^.N;  ready:= (Del=nil);

        while not ready do
        begin
          if (Del = nil) then
            ready := true
          else if (Del^.P.fyi > fmax) then
            begin
              ready:= true;  DelPrev^.N:= nil;
            end
          else
            begin
              DelPrev:= Del;  Del:= Del^.N;
            end;
        end;
      end;
    FreeAll(Del);
  end;
end;

global function Next (List: PairPtr) : PairPtr;   { Sets list pointer to the }
begin                                             { next list element        }
  Next:= List^.N;                                 {--------------------------}
end;

global function Head (List: PairPtr) : Pair;    { Delivers first pair of the }
begin                                           { list, i.e. the pair P with }
  Head:= List^.P;                               { the smallest value P.fyi.  }
end;                                            {----------------------------}

global procedure DelHead (var List: PairPtr);   { Deletes the first pair of  }
var                                             { the List.                  }
  Del : PairPtr;                                {----------------------------}
begin
  Del := List;  List:= List^.N;  Free(Del);
end;

global function Length (List: PairPtr) : integer;  { 'Length' delivers the   }
var  i : integer;                                  { number of elements in   }
begin                                              { list 'List'.            }
  i:= 0;                                           {-------------------------}
  while List <> nil do
  begin
    i:= succ(i); List:= List^.N;
  end;
  Length := i;
end;

{----------------------------------------------------------------------------}
{ Module initialization                                                      }
{----------------------------------------------------------------------------}
begin
  FreeList := nil;         { List of freed elements which can be used again  }
end.                       {-------------------------------------------------}
