{****************************************************************************}
{*                                                                          *}
{*                              Program/Module                              *}
{*                                   from                                   *}
{*                 NUMERICAL TOOLBOX FOR VERIFIED COMPUTING I               *}
{*                         Basic Numerical Problems                         *}
{*                                                                          *}
{*      Copyright (c) 1993   Rolf Hammer, Matthias Hocks, Dietmar Ratz      *}
{*                                                                          *}
{* This program/module is free software for non-commercial use. For details *}
{* on theory, algorithms, and programs, see the book                        *}
{*                                                                          *}
{*  R. Hammer, M. Hocks, U. Kulisch, D. Ratz:  Numerical Toolbox for        *}
{*  Verified Computing I - Basic Numerical Problems. Springer-Verlag,       *}
{*  Heidelberg (ISBN 3-540-57118-3), New York (ISBN 0-387-57118-3), 1993.   *}
{*                                                                          *}
{* This program/module is distributed WITHOUT ANY WARRANTY. For details,    *}
{* see the "Disclaimer / Legal Matters" of the book (page iv).              *}
{*                                                                          *}
{****************************************************************************}
{----------------------------------------------------------------------------}
{ Purpose: Computing enclosures for all solutions of systems of nonlinear    }
{    equations given by continuously differentiable functions.               }
{ Method: Extended interval Newton Gauss-Seidel method.                      }
{    Note that some routines are implemented using the "modified call by     }
{    reference" of PASCAL-XSC (see Language Reference for details) to avoid  }
{    the very inefficient allocation of local memory for the copies of the   }
{    actual parameters.                                                      }
{ Global procedures and functions:                                           }
{    procedure AllNLSS(...)       : computes enclosures for all solutions    }
{    function  AllNLSSErrMsg(...) : delivers an error message text           }
{ Updates:                                                                   }
{    07.03.1994 : Use of Epsilon in 'NewtonStep', bisection now in XINewton  }
{    05.05.1994 : Use of InnerIncl flag to decide whether to bisect the box  }
{    12.06.1995 : No 'p:=0' if NoSolution at the end of NewtonStep           }
{    23.11.1995 : Termination criterium changed to ' <= Epsilon'             }
{----------------------------------------------------------------------------}
module nlss;

use
  i_ari,      { Interval arithmetic                 }
  xi_ari,     { Extended interval arithmetic        }
  grad_ari,   { Gradient differentiation arithmetic }
  mv_ari,     { Real matrix/vector arithmetic       }
  mvi_ari,    { Interval matrix/vector arithmetic   }
  b_util,     { Boolean utilities                   }
  i_util,     { Interval utilities                  }
  mvi_util,   { Interval matrix/vector utilities    }
  matinv;     { Matrix inversion                    }

{----------------------------------------------------------------------------}
{ Error codes used in this module.                                           }
{----------------------------------------------------------------------------}
const
  NoError       = 0;  { No error occurred.                                   }
  lbSoluVecNot1 = 1;  { Lower bound of variable SoluVector is not equal to 1.}
  lbInfoVecNot1 = 2;  { Lower bound of variable InfoVector is not equal to 1.}
  VecsDiffer    = 3;  { Bounds of SoluVector and InfoVector do not match.    }
  VecTooSmall   = 4;  { SoluVector too small. Not all zeros can be stored.   }

{----------------------------------------------------------------------------}
{ Error messages depending on the error code.                                }
{----------------------------------------------------------------------------}
global function AllNLSSErrMsg ( Err : integer ) : string;
var
  Msg : string;
begin
  case Err of
    NoError       : Msg := '';
    lbSoluVecNot1 : Msg := 'Lower bound of SoluVector is not equal to 1';
    lbInfoVecNot1 : Msg := 'Lower bound of InfoVector is not equal to 1';
    VecsDiffer    : Msg := 'Bounds of SoluVector and InfoVector do not match';
    VecTooSmall   : Msg := 'Not all zeros found. SoluVector is too small';
    else          : Msg := 'Code not defined';
  end;
  if (Err <> NoError) then Msg := 'Error: ' + Msg + '!';
  AllNLSSErrMsg := Msg;
end;

{----------------------------------------------------------------------------}
{ Check if every component of interval vector 'iv' contains the value 'n'    }
{----------------------------------------------------------------------------}
operator in (n: integer; var iv: ivector) ResIn : boolean;
var
  k : integer;
begin
  k:= lb(iv);
  while (k < ub(iv)) and (n in iv[k]) do k:= k+1;
  ResIn:= (n in iv[k]);
end;

{----------------------------------------------------------------------------}
{ Purpose: Execution of one single interval Newton Gauss-Seidel step for the }
{    interval vector 'Y' and the function 'f'.                               }
{ Parameters:                                                                }
{    In     : 'f'          : must be declared for the type 'GTvector' to     }
{                            enable the internal use of the differentiation  }
{                            arithmetic 'grad_ari'.                          }
{             'Y'          : starting interval.                              }
{             'JfY'        : Jacobian matrix of 'f(Y)', already computed     }
{                            outside of 'NewtonStep'.                        }
{             'Epsilon'    : desired relative accuracy                       }
{    Out    : 'V'          : enclosures 'V[i]' for the splittings            }
{                            generated by the Newton step.                   }
{             'p'          : number of non-empty interval vectors 'V[i]'.    }
{ Description:                                                               }
{    'NewtonStep' executes the extended interval Newton Gauss-Seidel step    }
{    for 'Y' with result interval vector(s) 'V[i]' which can be empty.       }
{    'p' gives the number of non-empty interval vectors stored as vectors    }
{    'V[1]',...,'V[p]'.                                                      }
{----------------------------------------------------------------------------}
procedure NewtonStep (function f (x: GTvector) : GTvector[1..ub(x,1),
                                                          0..ub(x,2)];
                                             Y : ivector;
                           var             JfY : imatrix;
                                       Epsilon : real;
                           var               V : imatrix;
                           var               p : integer);
var
  c                     : rvector[1..ub(Y)];
  fC, b, Y_minus_c      : ivector[1..ub(Y)];
  R                     : rmatrix[1..ub(Y),1..ub(Y)];
  A                     : imatrix[1..ub(Y),1..ub(Y)];
  i, i0, n, InvErr, j   : integer;
  NoSolution            : boolean;
  h                     : interval;
  z                     : ivector[1..2];
begin
  c:= mid(Y);  fEvalJ(f, intval(c), fC);    { Midpoint evaluation of 'f'     }
                                            {--------------------------------}
  MatInv(mid(JfY),R,InvErr);                { Invert the midpoint matrix     }
  if InvErr <> 0 then R:= id(R);            {--------------------------------}

  A:= R * JfY;  b:= R * fC;             { Compute data for Gauss-Seidel step }
  Y_minus_c:= Y - c;                    {------------------------------------}

  p:= 0; i:= 0; i0:= 0; n:= ub(Y); { Initializations, A[i0,i0] contains zero }
  NoSolution:= false;              {-----------------------------------------}
  while (i < n) and (not NoSolution) do     { Interval Gauss-Seidel step for }
  begin                                     { non-zero A[i,i] elements       }
    i:= succ(i);                            {--------------------------------}
    if not (0 in A[i,i]) then
      begin
        if (RelDiam(Y[i]) > Epsilon) then
        begin
          h:= ## (b[i] + (for j:=1 to i-1 sum (A[i,j] * Y_minus_c[j]) )
                       + (for j:=i+1 to n sum (A[i,j] * Y_minus_c[j]) ) );
          h:= c[i] - h / A[i,i];

          if (Y[i] >< h) then
            NoSolution:= true
          else
            begin
              Y[i]:= Y[i] ** h;  Y_minus_c[i]:= Y[i] - c[i];
            end;
        end
      end
    else
      i0:= i;                                   { Largest i with 0 in A[i,i] }
  end;                                          {----------------------------}

  i:= 0;
  while (not NoSolution) and (i < i0) do    { Interval Gauss-Seidel step for }
  begin                                     { zero A[i,i] elements           }
    i:= succ(i);                            {--------------------------------}
    if (0 in A[i,i]) and (RelDiam(Y[i]) > Epsilon) then
    begin
      h:= ## ( b[i] + ( for j:=1 to i-1 sum (A[i,j] * Y_minus_c[j]) )
                    + ( for j:=i+1 to n sum (A[i,j] * Y_minus_c[j]) ) );
      z:= Y[i] ** (c[i] - h div A[i,i]);        { Extended interval division }
                                                {----------------------------}
      if (z[1] = EmptyIntval) then              { z[1] = z[2] = EmptyIntval  }
        NoSolution:= true                       {----------------------------}
      else
        begin                                              { Compute new 'Y' }
          Y[i]:= z[1];  Y_minus_c[i]:= Y[i] - c[i];        {-----------------}
          if z[2] <> EmptyIntval then
          begin                                   { Store further splittings }
            p:= p+1; V[p]:= Y; V[p][i]:= z[2];    {--------------------------}
          end
        end;
    end; { if 0 in A[i,i] ... }
  end; { while (not NoSolution) ... }

  if not NoSolution then
  begin
    p:= p+1;  V[p]:= Y;
  end;
end; { NewtonStep }

{----------------------------------------------------------------------------}
{ Purpose: Recursive procedure for the execution of the extended interval    }
{    Newton Gauss_Seidel method for the function 'f'.                        }
{ Parameters:                                                                }
{    In     : 'f'          : must be declared for the type 'GTvector' to     }
{                            enable the internal use of the differentiation  }
{                            arithmetic 'grad_ari'.                          }
{             'y'          : starting interval.                              }
{             'Epsilon'    : desired relative accuracy.                       }
{                            (interval diameter) of the result intervals.    }
{             'yUnique'    : signals whether it is already verified that the }
{                            actual interval 'y' contains a unique zero.     }
{    Out    : 'SoluVector' : enclosures for the zeros of 'f'.                }
{             'InfoVector' : corresponding information on the uniqueness of  }
{                            the zero in each of these enclosures.           }
{    In/Out : 'SoluNo'     : represents the number of the zero computed last }
{                            (in) and the total number of enclosures         }
{                            computed at the end of the recursion (out).     }
{ Description:                                                               }
{    The procedure 'XINewton' is recursively called whenever the extended    }
{    interval Newton step results in a splitting of the actual interval 'y'  }
{    in two or more intervals, and the tolerance condition is not fulfilled  }
{    yet. Otherwise, the enclosures for the zeros of 'f' are stored in the   }
{    interval matrix 'SoluVector' row by row, the corresponding information  }
{    on the uniqueness of the zero in each of these enclosures is stored in  }
{    the Boolean vector 'InfoVector'. The number of enclosures computed is   }
{    returned in the variable 'SoluNo'.                                      }
{----------------------------------------------------------------------------}
procedure XINewton (function f (x: GTvector) : GTvector[1..ub(x,1),
                                                        0..ub(x,2)];
                                           y : ivector;
                         var         Epsilon : real;
                                     yUnique : boolean;
                         var      SoluVector : imatrix;
                         var      InfoVector : bvector;
                         var          SoluNo : integer);
var
  fy         : ivector[1..ub(y)];
  dY         : rvector[1..ub(y)];
  Jfy        : imatrix[1..ub(y),1..ub(y)];
  yp         : imatrix[1..ub(y)+1,1..ub(y)];
  i, p, k, s : integer;
  my, supy   : real;
  InnerIncl  : boolean;
begin
  fJEvalJ(f, y, fy, Jfy);              { Compute f(y) and Jf(y)              }
  if 0 in fy then                      { Start if 0 in f(y), else do nothing }
  begin                                {-------------------------------------}
    NewtonStep(f,y,Jfy,Epsilon,yp,p);  { Extended interval Newton step with  }
                                       { results yp[i]                       }
    if (p = 1) then                    {-------------------------------------}
      begin
        InnerIncl := (yp[1] in y);         { Inner inclusion ===> uniqueness }
        yUnique := yUnique or InnerIncl;   {---------------------------------}
      end
    else
      begin
        InnerIncl := false;  yUnique := false;
      end;

    for i:=1 to p do
    begin
      if (MaxRelDiam(yp[i]) <= Epsilon) then
        begin { No more Newton steps }
          fEvalJ(f, yp[i], fy);                  { Compute f(yp[i])          }
          if (0 in fy) then                      { Store enclosure and info  }
          begin                                  {---------------------------}
            SoluNo := SoluNo + 1;
            if (SoluNo <= ub(SoluVector)) then
            begin
              SoluVector[SoluNo] := yp[i];     { Store enclosure of the zero }
              InfoVector[SoluNo] := yUnique;   { Store uniqueness info       }
            end;                               {-----------------------------}
          end;
        end
      else { Recursive call of 'XINewton' for interval 'yp[i]' }
        begin
          if not InnerIncl then                             { Bisect the box }
          begin                                             {----------------}
            yUnique := false;                               { Changed 5.5.94 }
            dY := diam(yp[i]);                              {----------------}
            s:= 1;
            for k:=2 to ub(Y) do
              if dY[k] > dY[s] then s:= k;
            my := mid(yp[i,s]);                 { Compute midpoint           }
            supy := yp[i,s].sup;                { Save supremum of component }
            yp[i,s].sup := my;                  { Set supremum to 'm'        }
                                                {----------------------------}
            XINewton(f,yp[i],Epsilon,yUnique,        { Call 'XINewton' for   }
                     SoluVector,InfoVector,SoluNo);  { first half of 'yp[i]' }
                                                     {-----------------------}
            yp[i,s].sup := supy;               { Reset supremum to old value }
            yp[i,s].inf := my;                 { Set infimum to 'm'          }
          end;                                 {-----------------------------}
          XINewton(f,yp[i],Epsilon,yUnique,       { Call 'XINewton' for      }
                   SoluVector,InfoVector,SoluNo); { (second half of) 'yp[i]' }
        end;                                      {--------------------------}
    end;
  end;
end;

{----------------------------------------------------------------------------}
{ Purpose: Execution of a verification step including the use of an epsilon  }
{    inflation.                                                              }
{ Parameters:                                                                }
{    In     : 'f'       : function of type 'GTvector'.                       }
{    Out    : 'yUnique' : returns 'true' if the verification is successful.  }
{    In/Out : 'y'       : interval enclosure to be verified.                 }
{ Description: This procedure checks the uniqueness of the zero enclosed in  }
{    the variable 'y' by an additional verification step including the use   }
{    of an epsilon inflation of the iterates.                                }
{----------------------------------------------------------------------------}
procedure VerificationStep (function f (x: GTvector) : GTvector[1..ub(x,1),
                                                                0..ub(x,2)];
                                 var               y : ivector;
                                 var         yUnique : boolean);
const
  kmax = 10;
var
  yIn, fY, yOld : ivector[1..ub(y)];
  JfY           : imatrix[1..ub(y),1..ub(y)];
  yp            : imatrix[1..ub(y)+1,1..ub(y)];
  k, p          : integer;
  eps           : real;
begin
  yIn := y;  k := 0;  eps:= 0.25;                          { Initializations }
  yUnique := false;                                        {-----------------}
  while (not yUnique) and (k < kmax) do { Do kmax loops to achieve inclusion }
  begin                                 {------------------------------------}
    yOld := blow(y,eps);                { Epsilon inflation of 'y'           }
                                        {------------------------------------}
    k := k+1;  fJEvalJ(f, yOld, fY, JfY);     { Perform interval Newton step }
    NewtonStep(f, yOld, JfY, 0, yp, p);       {------------------------------}

    if (p <> 1) then                              { No verification possible }
      k := kmax                                   {--------------------------}
    else if (yp[1] = yOld) then                   { Increase 'eps'           }
      eps := eps * 8                              {--------------------------}
    else
      begin
        y:= yp[1];  yUnique:= y in yOld;   { Inner inclusion ===> uniqueness }
      end;                                 {---------------------------------}
  end;
  if not yUnique then y:= yIn;
end;

{----------------------------------------------------------------------------}
{ Purpose: Computation of enclosures for all zeros of a continuously         }
{    differentiable multi-dimensional, vector-valued function.               }
{ Parameters:                                                                }
{    In     : 'f'             : objective function, must be declared for the }
{                               'GTvector' to enable the internal use of     }
{                               the differentiation arithmetic 'grad_ari'.   }
{             'Start',        : starting interval.                           }
{             'Epsilon'.      : desired relative accuracy                    }
{                               (interval diameter) of the result intervals. }
{    Out    : 'SoluVector'    : stores and returns the enclosures for the    }
{                               zeros of 'f'.                                }
{             'InfoVector'    : corresponding information on the uniqueness  }
{                               of the zeros in each of these enclosures.    }
{             'NumberOfSolus' : number of enclosures computed.               }
{             'Err'           : error code.                                  }
{ Description:                                                               }
{    The enclosures for the zeros of 'f' are computed by calling procedure   }
{    'XINewton'. Then an additional verification step is applied to those    }
{    enclosures which have not been verified.                                }
{    If an error occurs, the value of 'Err' is different from 0.             }
{----------------------------------------------------------------------------}
global procedure AllNLSS (function f (x: GTvector) : GTvector[1..ub(x,1),
                                                              0..ub(x,2)];
                                             Start : ivector;
                                           Epsilon : real;
                               var      SoluVector : imatrix;
                               var      InfoVector : bvector;
                               var   NumberOfSolus : integer;
                               var             Err : integer);

var
  i          : integer;
  MinEpsilon : real;
  Start1n    : ivector[1..ub(Start)-lb(Start)+1];
begin
  NumberOfSolus:= 0;
  if (lb(SoluVector) <> 1) then       { Check index bounds of result vectors }
    Err:= lbSoluVecNot1               {--------------------------------------}
  else if (lb(InfoVector) <> 1) then
    Err:= lbInfoVecNot1
  else if (ub(InfoVector) <> ub(SoluVector)) then
    Err:= VecsDiffer
  else
    begin                            { Start extended interval Newton method }
      Err:= NoError;                 {---------------------------------------}
      MinEpsilon:= succ(1.0) - 1.0;  { Relative machine accuracy (1 ulp)     }
      Start1n:= Start;               { Resize to standard bounds 1..n        }
                                     {---------------------------------------}
      if (Epsilon < MinEpsilon) then Epsilon := MinEpsilon;  { Set 'Epsilon' }
                                                             { to 1 ulp acc. }
      XINewton(f, Start1n, Epsilon, false,                   {---------------}
                  SoluVector, InfoVector, NumberOfSolus);

                                             { Check if there are more zeros }
      if ub(SoluVector) < NumberOfSolus then { than storage space            }
      begin                                  {-------------------------------}
        Err:= VecTooSmall;  NumberOfSolus:= ub(SoluVector);
      end;

      for i:=1 to NumberOfSolus do    { Verification step for the enclosures }
        if InfoVector[i] <> true then {--------------------------------------}
          VerificationStep(f,SoluVector[i],InfoVector[i]);
    end;
end;

{----------------------------------------------------------------------------}
{ Module initialization part                                                 }
{----------------------------------------------------------------------------}
begin
  { Nothing to initialize }
end.
